from chope.element import Element


class a(Element):
    pass


class abbr(Element):
    pass


class acronym(Element):
    pass


class address(Element):
    pass


class applet(Element):
    pass


class area(Element):
    pass


class article(Element):
    pass


class aside(Element):
    pass


class audio(Element):
    pass


class b(Element):
    pass


class base(Element):
    pass


class basefont(Element):
    pass


class bdi(Element):
    pass


class bdo(Element):
    pass


class big(Element):
    pass


class blockquote(Element):
    pass


class body(Element):
    pass


class br(Element):
    pass


class button(Element):
    pass


class canvas(Element):
    pass


class caption(Element):
    pass


class center(Element):
    pass


class cite(Element):
    pass


class code(Element):
    pass


class col(Element):
    pass


class colgroup(Element):
    pass


class data(Element):
    pass


class datalist(Element):
    pass


class dd(Element):
    pass


class details(Element):
    pass


class dfn(Element):
    pass


class dialog(Element):
    pass


class dir(Element):
    pass


class div(Element):
    pass


class dl(Element):
    pass


class dt(Element):
    pass


class em(Element):
    pass


class embed(Element):
    pass


class fieldset(Element):
    pass


class figcaption(Element):
    pass


class figure(Element):
    pass


class font(Element):
    pass


class footer(Element):
    pass


class form(Element):
    pass


class frame(Element):
    pass


class frameset(Element):
    pass


class h1(Element):
    pass


class h2(Element):
    pass


class h3(Element):
    pass


class h4(Element):
    pass


class h5(Element):
    pass


class h6(Element):
    pass


class head(Element):
    pass


class header(Element):
    pass


class hr(Element):
    pass


class html(Element):
    pass


class i(Element):
    pass


class iframe(Element):
    pass


class img(Element):
    pass


class input(Element):
    pass


class ins(Element):
    pass


class kbd(Element):
    pass


class label(Element):
    pass


class legend(Element):
    pass


class li(Element):
    pass


class link(Element):
    pass


class main(Element):
    pass


class map(Element):
    pass


class mark(Element):
    pass


class meta(Element):
    pass


class meter(Element):
    pass


class nav(Element):
    pass


class noframes(Element):
    pass


class noscript(Element):
    pass


class object(Element):
    pass


class ol(Element):
    pass


class optgroup(Element):
    pass


class option(Element):
    pass


class output(Element):
    pass


class p(Element):
    pass


class param(Element):
    pass


class picture(Element):
    pass


class pre(Element):
    pass


class progress(Element):
    pass


class q(Element):
    pass


class rp(Element):
    pass


class rt(Element):
    pass


class ruby(Element):
    pass


class s(Element):
    pass


class samp(Element):
    pass


class script(Element):
    pass


class section(Element):
    pass


class select(Element):
    pass


class small(Element):
    pass


class source(Element):
    pass


class span(Element):
    pass


class strike(Element):
    pass


class strong(Element):
    pass


class style(Element):
    pass


class sub(Element):
    pass


class summary(Element):
    pass


class sup(Element):
    pass


class svg(Element):
    pass


class table(Element):
    pass


class tbody(Element):
    pass


class td(Element):
    pass


class template(Element):
    pass


class textarea(Element):
    pass


class tfoot(Element):
    pass


class th(Element):
    pass


class thead(Element):
    pass


class time(Element):
    pass


class title(Element):
    pass


class tr(Element):
    pass


class track(Element):
    pass


class tt(Element):
    pass


class u(Element):
    pass


class ul(Element):
    pass


class var(Element):
    pass


class video(Element):
    pass


class wbr(Element):
    pass



import pytest
from chope import Element
from chope.css import Css

expected = \
"""<e1 class="my-class" color="yellow" size=123>
    text
    <e2>
        word
    </e2>
    <e3>
        letter
        space
    </e3>
</e1>"""


def test_should_render_nested_components_correctly():
    class e1(Element):
        pass

    class e2(Element):
        pass

    class e3(Element):
        pass

    component = e1(class_='my-class', color='yellow', size=123)[
        'text',
        e2[
            'word'
        ],
        e3[
            'letter',
            'space'
        ]
    ]

    assert component.render(4) == expected

def test_zero_indent_should_render_flat_string():
    expected = "<a>text</a>"

    class a(Element):
        pass

    assert a['text'].render(0) == expected

def test_when_negative_number_is_passed_to_render_should_render_with_zero_indent():
    class a(Element):
        pass

    assert a['text'].render(-1) == a['text'].render(0)

def test_able_to_render_css():
    expected = '<a>\n  b {\n    prop: text;\n  }\n</a>'

    class a(Element):
        pass

    comp = a[
        Css[
            'b': dict(prop='text')
        ]
    ]

    assert comp.render(2) == expected

def test_infer_id_and_classes_through_css_selector():
    expected = '<a id="id" class="class1 class2 class3 class4">text</a>'

    class a(Element):
        pass

    assert a('#id.class1.class2', class_='class3 class4')['text'].render(0) == expected

def test_should_raise_exception_if_id_detected_in_both_kwargs_and_css_selector():
    class a(Element):
        pass

    with pytest.raises(ValueError):
        a('#a', id='a')



from chope.css import Css, px, rem, percent, in_

expected = '''h1 {
   color: red;
   font-size: 1.2rem;
   padding: 1in;
   margin: 2%;
   outline: 1px dotted green;
}

.my-class {
   background: black;
}'''


def test_should_render_css_correctly():
    style = Css[
        'h1': dict(
            color='red',
            font_size=rem/1.2,
            padding=in_/1,
            margin=percent/2,
            outline=(px/1, 'dotted', 'green')
        ),
        '.my-class': dict(
            background='black'
        )
    ]

    assert style.render(3) == expected

def test_when_indent_is_zero_should_render_flat_string():
    expected = 'a {b: c;}d {e: f;}'

    style = Css['a': dict(b='c'), 'd': dict(e='f')]

    assert style.render(0) == expected



from typing import Iterable, List, Union


class Rule:
    def __init__(self, name: str, declarations: List[dict]):
        self.__declarations = declarations
        self.__name = name

    def render(self, indent: int = 2) -> str:
        nl = '\n'
        indented = indent > 0
        declarations_str = ''
        for property, value in self.__declarations.items():
            if isinstance(value, Iterable) and not isinstance(value, str):
                value = ' '.join(value)

            declarations_str += \
                f'{" " * indent}{property.replace("_", "-")}: {value};{nl * indented}'

        return f'{self.__name} {{{nl * indented}{declarations_str}}}'


class Css:
    def __init__(self, rules: List[Rule]):
        self._rules = rules

    def __class_getitem__(cls, items: Union[slice, Iterable[slice]]) -> 'Css':
        if isinstance(items, slice):
            rules = [Rule(items.start.replace('_', '-'), items.stop)]
        else:
            rules = [Rule(item.start.replace('_', '-'), item.stop)
                     for item in items]

        return cls(rules)

    def render(self, indent: int = 2) -> str:
        indented = indent > 0
        return ('\n\n' * indented).join((rule.render(indent) for rule in self._rules))


class Unit:
    def __init__(self, name: str):
        self.__name = name

    def __truediv__(self, value) -> str:
        return str(value) + self.__name


cm = Unit('cm')
ch = Unit('ch')
em = Unit('em')
ex = Unit('ex')
in_ = Unit('in')
mm = Unit('mm')
pc = Unit('pc')
percent = Unit('%')
pt = Unit('pt')
px = Unit('px')
rem = Unit('rem')
vh = Unit('vh')
vmax = Unit('vmax')
vmin = Unit('vmin')
vw = Unit('vw')



import re
from typing import Iterable, Union

from chope.css import Css


class Element:
    def __init__(self, *args, **kwargs):
        self._components: Iterable[Component] = []
        self._attributes = kwargs
        self._classes = self._attributes.pop('class_', '')
        self._id = self._attributes.pop('id', '')

        if args:
            selector_pattern = \
                r'^(?:#([^\s\.#]+))?(?:\.([^\s#]+))?'  ## #id.class1.class2
            
            id, classes = re.findall(selector_pattern, args[0])[0]

            if id and self._id:
                raise ValueError(
                    f'id declared twice: #{id} and id="{self._id}"')

            self._id = id
            self._classes = f'{classes.replace(".", " ")} {self._classes}'.strip()

    def __class_getitem__(cls, comps: Union['Component', Iterable['Component']]) \
            -> 'Element':
        inst = cls()
        if isinstance(comps, Component.__args__):
            inst._components = (comps,)
        else:
            inst._components = comps

        return inst

    def __getitem__(self, comps: Union['Component', Iterable['Component']]) \
            -> 'Element':
        if isinstance(comps, Component.__args__):
            self._components = (comps,)
        else:
            self._components = comps

        return self

    def render(self, indent: int = 2) -> str:
        nl = '\n'
        indented = indent > 0

        comp_str = nl * indented
        for comp in self._components:
            if isinstance(comp, str):
                comp_str += f'{" " * indent}{comp}{nl * indented}'
            else:
                _comp_str = comp.render(indent).replace(
                    nl, f'{nl * indented}{" " * indent}')
                comp_str += f'{" " * indent}{_comp_str}{nl * indented}'

        name = self.__class__.__name__

        attrs_str = f' id="{self._id}"' if self._id else ''
        attrs_str += f' class="{self._classes}"' if self._classes else ''

        for attr, val in self._attributes.items():
            val = f'"{val}"' if isinstance(val, str) else str(val)

            attrs_str += f' {attr}={val}'

        return f'<{name}{attrs_str}>{comp_str}</{name}>'


Component = Union[str, Element, Css]
